<?php

namespace Tangible\LifterElements;

defined( 'ABSPATH' ) or die();

use BetterLifter as lift;
use Tangible\LifterElements\Utils as utils;

/**
 * Main LifterElements Class: Handle the logic with Elementor
 *
 * @see https://developers.elementor.com/creating-an-extension-for-elementor/
 * 
 * @since 1.0.0
 */
final class Elementor {

	/**
	 * Plugin Version
	 *
	 * @since 1.0.0
	 *
	 * @var string The plugin version.
	 */
	const VERSION = LifterElements_VER;

	/**
	 * Minimum Elementor Version
	 *
	 * @since 1.0.0
	 *
	 * @var string Minimum Elementor version required to run the plugin.
	 */
	const MINIMUM_ELEMENTOR_VERSION = '2.0.0';

	/**
	 * Minimum PHP Version
	 *
	 * @since 1.0.0
	 *
	 * @var string Minimum PHP version required to run the plugin.
	 */
	const MINIMUM_PHP_VERSION = '7.0';

	/**
	 * Instance
	 *
	 * @since 1.0.0
	 *
	 * @access private
	 * @static
	 *
	 * @var LifterElement The single instance of the class.
	 */
	private static $_instance = null;

  /**
   * @see ./Plugin.php
   */
  public $plugin;

	/**
	 * Instance
	 *
	 * Ensures only one instance of the class is loaded or can be loaded.
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 * @static
	 *
	 * @return LifterElement An instance of the class.
	 */
	public static function instance() {

		if ( is_null( self::$_instance ) ) {
			self::$_instance = new self();
		}

		return self::$_instance;
	}

	/**
	 * Constructor
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 */
	public function __construct() {

		add_action( 'init', [ $this, 'i18n' ] );

		// Add our custom category
		add_action( 'elementor/elements/categories_registered', [ $this, 'add_category' ] );
	}

	/**
	 * Load Textdomain
	 *
	 * Load plugin localization files.
	 *
	 * Fired by `init` action hook.
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 */
	public function i18n() {
		load_plugin_textdomain( 'lifter-elements' );
	}

	/**
	 * Initialize the plugin
	 *
	 * Load the plugin only after Elementor (and other plugins) are loaded.
	 * Checks for basic plugin requirements, if one check fail don't continue,
	 * if all check have passed load the files required to run the plugin.
	 *
	 * Fired by `plugins_loaded` action hook.
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 */
	public function init( $plugin ) {

		// Required for the widget and dynamic tags definition
		if ( $plugin ) $this->plugin = $plugin;

		// Handle the logic specific to the elementor editor
		new Editor();

		// Check if Elementor installed and activated
		if ( !did_action( 'elementor/loaded' ) ) {
			add_action( 'admin_notices', [ $this, 'admin_notice_missing_main_plugin' ] );
			return;
		}

		// Check for required Elementor version
		if ( !version_compare( ELEMENTOR_VERSION, self::MINIMUM_ELEMENTOR_VERSION, '>=' ) ) {
			add_action( 'admin_notices', [ $this, 'admin_notice_minimum_elementor_version' ] );
			return;
		}

		// Check for required PHP version
		if ( version_compare( PHP_VERSION, self::MINIMUM_PHP_VERSION, '<' ) ) {
			add_action( 'admin_notices', [ $this, 'admin_notice_minimum_php_version' ] );
			return;
		}

		// Add Plugin actions
		add_action( 'elementor/widgets/register', [ $this, 'init_widgets' ] );
		add_action( 'elementor/dynamic_tags/register', [ $this, 'init_tags' ] );
		add_action( 'elementor/controls/register', [ $this, 'init_controls' ] );

		// Register Frontend Styles
		add_action( 'elementor/frontend/after_enqueue_styles', [ $this, 'frontend_styles' ] );
		add_action( 'elementor/frontend/after_register_scripts', [ $this, 'frontend_scripts' ] );

    // Handle Lesson page drip restriction
    add_action( 'wp', [ $this, 'dripped_lesson_not_restricted'] );

    // Handle Course Syllabus Widget lesson-preview template part drip restriction
    add_action( 'remove_lesson_preview_drip_restriction', [ $this, 'dripped_lesson_not_restricted'] );
	}

  /**
   * Handle Dripped Lesson view restriction
   *
   * @param  object $lesson
   * @return void
   *
   */
  public function dripped_lesson_not_restricted( $lesson = null ) {

    if ( utils\is_builder() || utils\is_preview() ) return;

    // if not lesson object passed get it from current post
    if( !$lesson  || !($lesson instanceof \LLMS_Lesson) ) $lesson = lift\lesson_lifter();

    // if still not lesson return
    if( !$lesson ) return;

    // check if dripped lesson view allowed - if not return
    $allowed = lift\is_dripped_lesson_view_allowed( $lesson );
    if( !$allowed ) return;

    /**
     * Remove Lesson drip restriction if allowed
     *
     *  @see in lifterlms\includes\functions\llms.functions.access.php
     *
     *  apply_filters( 'llms_page_restricted', $results, $post_id )
     */
    add_filter( 'llms_page_restricted', function( $results, $post_id ) {

      if( isset($results['reason']) && $results['reason'] === 'lesson_drip' ) {
        $results['is_restricted'] = 0;
      }

      return $results;
    }, 10, 2 );
  }

	/**
	 * 	Register custom CSS
	 *
	 *  @since 1.0.0
	 *
	 * @access public
	 */
	public function frontend_styles() {
		wp_enqueue_style( 'le-frontend', LifterElements_URL . 'assets/build/frontend.min.css', [], LifterElements_VER );
	}

	/**
	 * 	Register custom JS/lib
	 *
	 *  @since 1.0.0
	 *
	 * @access public
	 */
	public function frontend_scripts() {

		// Needed for circular gradient in progress widget
		wp_register_script( 'conic-gradient', LifterElements_URL . 'assets/lib/conic-gradient.js' );

		// Widgets
		wp_enqueue_script( 'le-frontend-js', LifterElements_URL . 'assets/build/frontend.min.js', ['jquery'], LifterElements_VER );

		$el_version = defined('ELEMENTOR_PRO_VERSION')
			? ELEMENTOR_PRO_VERSION
			: (defined('ELEMENTOR_VERSION') ? ELEMENTOR_VERSION : '0')
		;

		// If config is needed by the JavaScript
		wp_localize_script( 'le-frontend-js', 'LE_Widgets', [
      'course_access_plans' => [
        'is_legacy' => version_compare($el_version, '2.7.0') >= 0 ? 'no' : 'yes'
      ]
		]);

		wp_localize_script( 'le-frontend-js', 'LE_Driptimer', [
			'time_labels' => [
				'days'    => __( 'Days', 'lifter-elements' ),
				'hours'   => __( 'Hours', 'lifter-elements' ),
				'minutes' => __( 'Minutes', 'lifter-elements' ),
				'seconds' => __( 'Seconds', 'lifter-elements' ),
			]
		]);
	}

	/**
	 * Admin notice
	 *
	 * Warning when the site doesn't have Elementor installed or activated.
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 */
	public function admin_notice_missing_main_plugin() {

		if ( isset( $_GET['activate'] ) ) unset( $_GET['activate'] );

		$message = sprintf(
			esc_html__( '"%1$s" requires "%2$s" to be installed and activated.', 'lifter-elements' ),
			'<strong>' . esc_html__( 'Lifter Elements', 'lifter-elements' ) . '</strong>',
			'<strong>' . esc_html__( 'Elementor', 'lifter-elements' ) . '</strong>'
		);

		printf( '<div class="notice notice-warning is-dismissible"><p>%1$s</p></div>', $message );
	}

	/**
	 * Admin notice
	 *
	 * Warning when the site doesn't have a minimum required Elementor version.
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 */
	public function admin_notice_minimum_elementor_version() {

		if ( isset( $_GET['activate'] ) ) unset( $_GET['activate'] );

		$message = sprintf(
			esc_html__( '"%1$s" requires "%2$s" version %3$s or greater.', 'lifter-elements' ),
			'<strong>' . esc_html__( 'Lifter Elements', 'lifter-elements' ) . '</strong>',
			'<strong>' . esc_html__( 'Elementor', 'lifter-elements' ) . '</strong>',
			 self::MINIMUM_ELEMENTOR_VERSION
		);

		printf( '<div class="notice notice-warning is-dismissible"><p>%1$s</p></div>', $message );
	}

	/**
	 * Admin notice
	 *
	 * Warning when the site doesn't have a minimum required PHP version.
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 */
	public function admin_notice_minimum_php_version() {

		if ( isset( $_GET['activate'] ) ) unset( $_GET['activate'] );

		$message = sprintf(
			/* translators: 1: Plugin name 2: PHP 3: Required PHP version */
			esc_html__( '"%1$s" requires "%2$s" version %3$s or greater.', 'lifter-elements' ),
			'<strong>' . esc_html__( 'Lifter Elements', 'lifter-elements' ) . '</strong>',
			'<strong>' . esc_html__( 'PHP', 'lifter-elements' ) . '</strong>',
			 self::MINIMUM_PHP_VERSION
		);

		printf( '<div class="notice notice-warning is-dismissible"><p>%1$s</p></div>', $message );
	}

	/**
	 * Init Widgets
	 *
	 * Include widgets files and register them
	 *
	 * @since 1.0.0
	 * @see get_dynamic_tags()  method in inc/Plugin.php
	 * 
	 * @access public
	 */
	public function init_widgets() {
		
		// Register widgets
		$widgets_manager = &\Elementor\Plugin::instance()->widgets_manager;
		
		if( !isset($this->plugin) ) return;

		foreach( $this->plugin->widgets as $category ) {
			foreach( $category as $widget ) {

				// If the module extends a class, verify that the class is defined
				if( !isset($widget['parent']) || ($widget['parent'] && class_exists($widget['parent'])) ) {

					// Widget class's file
					require_once( $widget['path'] );

					// Register widget
					$widgets_manager->register( new $widget['namespace']() );

					// If we need to modify controls from the parent class
					if( isset($widget['parent']) ) $this->update_controls_widgets($widget);
				}
			}
		}
	}

	/**
	 * Trigger the modify_controls and the add_controls method of widgets (if exist) for allowing them
	 * to update the controls registered by the parent class
	 * 
	 * @param array $widget 	See init_widgets()
	 */
	private function update_controls_widgets( $widget ) {
		
		if( method_exists($widget['namespace'], 'add_controls') ) {
			add_action( 'elementor/element/before_section_start', function( $element, $section_id, $args ) use( $widget ) {
				if( utils\is_namespace_match_object($element, $widget['namespace']) ) {
					$widget['namespace']::add_controls( $element, $section_id, $args );
				}
			}, 9, 3 );
		}

		/**
		 * Modify existing controls
		 */
		if( method_exists($widget['namespace'], 'modify_controls') ) {
			add_action( 'elementor/element/before_section_end', function( $element, $section_id, $args ) use( $widget ) {
				if( utils\is_namespace_match_object($element, $widget['namespace']) ) {
					$widget['namespace']::modify_controls( $element, $section_id, $args );
				}
			}, 99, 3);
		}
	}

	/**
	 * Init Controls
	 *
	 * Include controls files and register them
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 */
	public function init_controls() {}

	/**
	 * Init Dynamic Tags
	 *
	 * @since 1.0.0
	 *
	 * @access public
	 */
	public function init_tags( $dynamic_tags ) {
		
    /**
		 * Note:
		 * 
		 * In the futur, it looks possible  to simplify the creation of dynamics tags by  
		 * using anonymous classes
		 */
		if( !isset($this->plugin)) return;
		
		foreach( $this->plugin->dynamic_tags as $category ) {
			
			\Elementor\Plugin::$instance->dynamic_tags->register_group( $category['group'], [
				'title' => 'LifterElements: ' . $category['label'] 
			]);
			
			foreach( $category['tags'] as $slug => $tag ) {

				// If the module extends a class, verify that the class is defined
				if( !isset($tag['parent']) || ($tag['parent'] && class_exists($tag['parent'])) ){

					// Tag class's file
					require_once( $tag['path'] );
          
					// Register tag
					$dynamic_tags->register( new $tag['namespace']() );
				}
				
			}
		}
	}

	/**
	 * Add a custom categroy
	 *
	 * @see https://developers.elementor.com/widget-categories/
	 */
	public function add_category( $elements_manager ) {

		$elements_manager->add_category(
			'lifter-elements',
			[
				'title' => __( 'LifterLMS - General', 'lifter-elements' ),
				'icon'  => 'fa fa-plug',
			]
		);
	}

}

Elementor::instance();
