<?php

namespace Tangible\LifterElements;

defined( 'ABSPATH' ) or die();

use Tangible\LifterElements\Utils as utils;

/**
 * Global controller of the plugin
 */
class Plugin {

	/**
	 * Path to the widgets folder
	 * 
	 * @var string
	 */
	private $widgets_path = LifterElements_DIR . 'includes/widgets/';


	/**
	 * Path to the dynamic-tags folder
	 * 
	 * @var string
	 */
	private $dynamic_tags_path = LifterElements_DIR . 'includes/dynamic-tags/';


	/**
	 * Namespace of the widgets
	 * 
	 * @var string
	 */
	private $widgets_namespace = '\Tangible\LifterElements\Widgets';
	

	/**
	 * Namespace of the dynamic tags
	 * 
	 * @var striner
	 */
	private $dynamic_tags_namespace = '\Tangible\LifterElements\DynamicTags';


	/**
	 * Array with the infos needed by Elementor for register our widgets
	 * 
	 * @var array
	 */
	public $widgets = array();

	
	/**
	 * Array with the infos needed by Elementor for register our dynamic tags
	 * 
	 * @var array
	 */
	public $dynamic_tags = array();


	/**
	 * Array with the infos needed for register our shortcodes
	 * 
	 * @var array
	 */
	public $shortcodes = array();

  /**
   * Object - plugin API instance returned from framework register plugin
   *
   * @var object
   */
	public $api;

	/**
	 * Constructor
	 */
	public function __construct() {

		// Get the widgets/tags/shortcodes we need to defined (from json files)
		$this->set_dynamic_tags();
		$this->set_widgets();
		$this->set_shortcodes();

		// Register shortcodes
		$this->register_shortcodes();

		// Initialize the tangible framework
		new Framework( $this, function() {
      // Include the group related file after the initialization for the framework
      require_once LifterElements_DIR . 'includes/groups/index.php';
    });

		// Handle the elementor logic
		$elementor = new Elementor();

		// Admin's hooks
		if( is_admin() ) $this->admin_hooks();

    $this->extensions();

		add_action( 'plugins_loaded', function() use( $elementor ) {

      // Plugin settings
      $settings = $this->api->get_settings();

      // Handle Dripped Lesson view allowed settings
      add_filter( 'betterlifter_lesson_settings', function( $categories ) use( $settings ) {
        
        // settings category name
        $cat_name = 'lessons';
        
        // course categories for dripped lesson view allowed field name
        $field_name_c = 'drip_view_course_categories';

        // all lessons dripped view allowed field name
        $field_name_l = 'drip_view_all_lessons';

        // all lessons dripped view allowed settings value
        $drip_view_all = (isset($settings[$cat_name][$field_name_l])
          && $settings[$cat_name][$field_name_l] ==='true' )
          ? 'true'
          : 'false';

        // course categories for dripped lesson view allowed settings value
        $drip_view_categories = (isset($settings[$cat_name][$field_name_c])
          &&  $settings[$cat_name][$field_name_c] )
          ? $settings[$cat_name][$field_name_c]
          : '';
        
        $drip_lessons_settings = [
          'drip_view_all_lessons' => $drip_view_all,
          'drip_view_course_categories' => $drip_view_categories
        ];

        return array_merge( $categories, $drip_lessons_settings);
      }, 10, 1);

			// Pass data to the elementor instance
			$elementor->init( $this );

			// Allow us to create lessons and quizzes template 
			$this->post_type_support();
			
		}, 7); // priority set at 7 : before Elementor plugin - due to 'register_post_type_args', and after 6 when this plugin registers to tangible framework
  }


  /**
   * Initiate action on the admin side
   */
  private function admin_hooks() {

    add_action( 'admin_enqueue_scripts', function() {
      wp_enqueue_style( 'le-admin', LifterElements_URL . 'assets/build/admin.min.css' );
    }, 10, 1 );
  }

    
  /**
   * Load extensions logic
   */
  private function extensions() {
    require_once __DIR__ . '/extensions/index.php';
    require_once __DIR__ . '/widget-extensions/index.php';
  }


  /**
   * Add the possibility to create a template for lesson/quizz/assignements
   * by changing the param of the CPT
   */
  private function post_type_support() {
  
    add_filter( 'register_post_type_args', function( $args, $post_type ) {
        
      // LifterLMS post type (filter: some post-type are added only according to the  active extensions)
      $le_post_types = apply_filters( 'le-elementor-post-type', [
        "lesson", 
        "llms_assignment", 
        "llms_quiz"
      ]);

      if( in_array($post_type, $le_post_types) ) {
        $args['exclude_from_search'] = false;
        $args['show_in_nav_menus'] = true;
      }

      return $args;
    }, 20, 2 );

  }


  /**
   * Get the widgets config from the json file
   * 
   * @see ./congig/widgets.json
   */
  private function set_widgets() {

    $widgets_data = file_get_contents( LifterElements_DIR . 'config/widgets.json' );
    $widgets_data = json_decode( $widgets_data, true );

    // We will use this filter for defined widgets which needs an extenstion
    $widgets_data = apply_filters( 'le-elementor-register-widgets', $widgets_data );

    if(!empty($widgets_data)) {
      foreach( $widgets_data as $key => &$category ) {
        
        $category_path 		  = $this->widgets_path . $key . '/';
        $category_namespace = $this->get_namespace_by_widget_category($key);
        
        // Module by categories
        foreach( $category as $clef => &$widget ) {
          
          // convert to Pascal case for getting the class name
          $class_name = str_replace('-', '', ucwords($clef, '-'));
          
          $widget['path'] 		  = $category_path . $class_name . '.php';
          $widget['namespace']  = $category_namespace . '\\' . $class_name;
          
          if( !isset($widget['label']) )
            $widget['label'] = utils\kebabcase_to_text( $clef );
        }

      }
    }

    $this->widgets = !empty($widgets_data) ? $widgets_data : array();
  }


  /**
   * Get the dynamic_tags config from the json file
   * 
   * @see ./congig/dynamic_tags.json
   */
  private function set_dynamic_tags() {

    $dynamic_tags_data = file_get_contents( LifterElements_DIR . 'config/dynamic-tags.json' );
    $dynamic_tags_data = json_decode( $dynamic_tags_data, true );

    // We will use this filter for defined tags which needs an extenstion
    $dynamic_tags_data = apply_filters( 'le-elementor-register-tags', $dynamic_tags_data );

    if(!empty($dynamic_tags_data)) {
      foreach( $dynamic_tags_data as $key => &$category ) {
        
        $category_path 		  = $this->dynamic_tags_path . $key . '/';
        $category_namespace = $this->get_namespace_by_tag_category($key);
        
        // Prefix the group
        $category['group'] = "le-" . $category['group'];
        
        // Module by categories
        foreach( $category['tags'] as $clef => &$tag ) {
          
          // convert to Pascal case for getting the class name
          $class_name = str_replace('-', '', ucwords($clef, '-'));
          
          $tag['path']        = $category_path . $class_name . '.php';
          $tag['namespace'] 	= $category_namespace . '\\' . $class_name;

          if( !isset($tag['label']) )
            $tag['label'] = utils\kebabcase_to_text( $clef );
        }

      }
    }
    
    $this->dynamic_tags = !empty($dynamic_tags_data) ? $dynamic_tags_data : array();
  }


  /**
   * Get the shortcodes config from the json file
   * 
   * @see ./congig/shortcodes.json
   */
  private function set_shortcodes() {

    $shortcodes_data = file_get_contents( LifterElements_DIR . 'config/shortcodes.json' );
    $shortcodes_data = json_decode( $shortcodes_data, true );

    // We will use this filter for shortcodes which needs an extenstion
    $shortcodes_data = apply_filters( 'le-elementor-register-shortcodes', $shortcodes_data );

    if(!empty($shortcodes_data)) {
      foreach( $shortcodes_data as $key => &$category ) {
        
        // Module by categories
        foreach( $category['sections'] as $clef => &$section ) {
          foreach( $section as $slug => &$shortcode ) {

            // Shortcode [tag]
            $shortcode_data['tag']       = 'le-' . $key . '-' . $slug;
            $shortcode_data['function']  = 'BetterLifter\\' . $shortcode;
          
          $shortcode = $shortcode_data; 			    		
            
            if( !isset($shortcode['label']) )
              $shortcode['label'] = utils\kebabcase_to_text( $slug );
          }
        }

      }
    }
    
    $this->shortcodes = !empty($shortcodes_data) ? $shortcodes_data : array();
  }


  /**
   * Register all the shortcofde of the plugin
   */
  private function register_shortcodes() {

    add_action( 'init', function() {
      if( !empty($this->shortcodes) ) {
        foreach( $this->shortcodes as $category_slug => $category ) {
          foreach( $category['sections'] as $section ) {
            foreach( $section as $shortcode ) { 			
              // Register the shortcode 
              add_shortcode( $shortcode['tag'], $shortcode['function'] );
            }
          }
        }
      }
    });
  }


  /**
   * Return the folder path of the category name (widgets)
   * 
   * @param string $category Category name from the json config file
   * 
   * @return string
   */
  private function get_namespace_by_widget_category( $category ) {
    $category_namespace = str_replace('-', '', ucwords($category, '-'));	
    return $this->widgets_namespace . '\\' . $category_namespace;
  }


  /**
   * Return the folder path of the category name (tags)
   * 
   * @param string $category Category name from the json config file
   * 
   * @return string
   */
  private function get_namespace_by_tag_category( $category ) {
    $category_namespace = str_replace('-', '', ucwords($category, '-'));	
    return $this->dynamic_tags_namespace . '\\' . $category_namespace;
  }

}
