<?php

namespace Tangible\WordpressVideo;

/**
 * Manage video output for WordPress
 * Currently support direct link, Youtube and Vimeo links
 * 
 * Not that for Vimeo, some options can be not available 
 * according to the type of account 
 * 
 * This folder could be turn into a module later
 */
class Video {

  /**
  * Url to the video
  *
  * @var        string
  */
  public $url = '';


  /**
  * If set, contain the video object (Instance of Youtube, Vimeo or Link)  
  *
  * @var        mixed
  */
  private $video = false;


  /**
   * Type of video (host, direct link ... etc)
   *
   * @var        string
   */
  public $video_type = '';


  /**
  * Constructor
  *
  * @param      string  $url    The url
  * @param      array   $args   Arguments
  */
  public function __construct( $url, $args = array() ) {

    if( !filter_var($url, FILTER_VALIDATE_URL) ) return;
    
    $this->url = $url;
    $this->args = (array) $args;

    $this->set_video_type();
    $this->video = $this->get_video_object();
  }


  /**
  * Sets the video type.
  */
  private function set_video_type() {

    if( strpos($this->url, 'youtube.com') !== false || strpos($this->url, 'youtu.be') !== false ) {
      $this->video_type = 'youtube';
    }
    else if( strpos($this->url, 'vimeo.com') !== false ) {
      $this->video_type = 'vimeo';
    }
    else if( empty(wp_oembed_get( $this->url )) ) {
      $this->video_type = 'direct_link'; 
    }
  }


  /**
  * Gets the final HTML output.
  *
  * @return     string  The HTML output.
  */
  public function get_output() {

    if( empty($this->video) ) return '';

    return $this->video->get_output(); 
  } 


  /**
   * Return the right object accroding to the video class 
   *
   * @return     object Link|Vimeo|Youtube  The video object.
   */
  public function get_video_object() {

    if( $this->video !== false ) return $this->video;

    switch( $this->video_type ) {
        
      case 'youtube':
        return new Youtube( $this->url, $this->args );
        break;
      
      case 'vimeo':
        return new Vimeo( $this->url, $this->args );
        break;
      
      case 'direct_link': 
        return new Link( $this->url, $this->args );
        break;

      /**
       * If the link is not matching a provider compatible with the module, we check
       * if it's compatible with the wp_embed WordPress function.
       * 
       * If it is the case, we will be able to display the video but without all the options
       * (mute sound, autoplay ...etc) 
       */
      default:
        return new WPEmbed( $this->url, $this->args );
        break; 
    }

  }
}





