<?php

namespace Tangible\LifterElements\Widgets\Lessons;

defined( 'ABSPATH' ) or die();

use BetterLifter as lift;
use Tangible\LifterElements\Utils as utils;

/**
 * @see https://developers.elementor.com/creating-a-new-widget/
 */
class MarkCompleteButton extends \Elementor\Widget_Button {


	/**
	 * Widget slug
	 * 
	 * @return string
	 */
	public function get_name() {
		return 'le-lesson-mark-complete-button';
	}


	/**
	 * Widget name
	 * 
	 * @return string
	 */
	public function get_title() {
		return __( 'Mark Complete Lesson', 'lifter-elements' );
	}


	/**
	 * @see https://developers.elementor.com/widget-categories/
	 * 
	 * @return array
	 */
	public function get_categories() {
		return array( 'lifter-elements' );
	}


	/**
	 * Add before the existing controls 
	 *
	 * @see https://code.elementor.com/php-hooks/#elementorelementbefore_section_start
	 * @see inc/Elementor.php in the update_controls_widgets() method
	 * 
	 * @param      obj     $element     The element
	 * @param      string  $section_id  The section identifier
	 * @param      array   $args        The arguments
	 */
	public static function add_controls( $element, $section_id, $args ) {
		
		if( $section_id !== 'section_button'  ) return;

		$custom_condition = [
			'terms' => [
				[
					'name' 		 => 'render_type',
					'operator' => 'in',
					'value' 	 => [
						'custom_button'
					],
				],
			]
		];

		// Content section
		$element->start_controls_section(
			'link_type',
			[
				'label' => __( 'Content', 'lifter-elements' ),
				'tab' 	=> \Elementor\Controls_Manager::TAB_CONTENT,
			]
		);

		$element->add_control(
			'render_type', [
				'label' 	=> __( 'Render', 'lifter-elements' ),
				'type' 		=> \Elementor\Controls_Manager::SELECT,
				'default'	=> 'custom_button',
				'options'	=> [
					'lifter'       => __( 'LifterLMS render', 'lifter-elements' ),
					'custom_button'=> __( 'Custom button', 'lifter-elements' ),
				],
			]
		);

		$element->add_control(
			'hr', [
				'type'      => \Elementor\Controls_Manager::DIVIDER,
				'conditions'=> $custom_condition
			]
		);

		$element->start_controls_tabs( 'tabs_typo_style' );
		
		// Create a tab for each type of button
		foreach( ['complete', 'incomplete', 'quiz', 'assignment'] as $value ) {

			
			$element->start_controls_tab(
				"button_{$value}_tab",
				[
					'label'      => __( ucfirst($value), 'lifter-elements' ),
					'conditions' => $custom_condition
				]
			);

      // Explanation of how to make the "mark incomplete" button works
      if( $value === 'incomplete' ) {
        $element->add_control(
          'mark_incomplete_note',
          [
            'type'	=> \Elementor\Controls_Manager::RAW_HTML,
            'raw' 	=> '<i>' 
                        . __( 'The mark incomplete button will work only if the "Retake Lessons" setting is enabled in LifterLMS', 'lifter-elements' ) 
                        . ' (see <a href="' . \get_admin_url() . 'admin.php?page=llms-settings&tab=courses' . '">course settings</a>).' 
                      . '</i>',
          ]
        );
      } 

			// Default text
			if( in_array($value, ['incomplete', 'complete'] ) ) {
				$default =  __( 'Mark ' . ucfirst($value), 'lifter-elements' );
      }
      else if( $value === 'quiz' ) {
				$default =  __( 'Take Quiz', 'lifter-elements' );
      }
			else {
				$default =  __( 'Start Assignment', 'lifter-elements' );
      }
				
			$element->add_control(
				"text_{$value}", [
					'label' 		=> __( 'Text', 'lifter-elements' ),
					'type' 			=> \Elementor\Controls_Manager::TEXT,
					'default'		=> $default,
					'conditions'=> $custom_condition
				]
			);

			$element->add_control(
				"button_{$value}_color",
				[
					'label' 	  => __( 'Color', 'lifter-elements' ),
					'type' 		  => \Elementor\Controls_Manager::COLOR,
					'selectors' => [
						"{{WRAPPER}} .le-button-mark-{$value} a.elementor-button" => 'background-color: {{VALUE}}',
						"{{WRAPPER}} .le-button-mark-{$value} .elementor-button" => 'background-color: {{VALUE}}'
					],
					'conditions' => $custom_condition,
				]
			);


			$element->add_group_control(
				\Elementor\Group_Control_Typography::get_type(),
				[
					'name'       => "{$value}_typography",
					'label'      => __( 'Typography', 'lifter-elements' ),
					'scheme' 	   => \Elementor\Core\Schemes\Typography::TYPOGRAPHY_1,
					'selector' 	 => "{{WRAPPER}} div.le-button-mark-{$value} .elementor-button span.elementor-button-text",
					'conditions' => $custom_condition,
				]
			);	

			$element->end_controls_tab();
		}

		$element->end_controls_tabs();	

		$element->end_controls_section();

	}


	/**
	 * Modify control defined in the parent class
	 * 
	 * @see inc/Elementor.php in the update_controls_widgets() method
	 * 
	 * @param      obj     $element     The element
	 * @param      string  $section_id  The section identifier
	 * @param      array   $args        The arguments
	 */
	public static function modify_controls( $element, $section_id, $args ) {

		// Hide the controls if it's not a custom button 
		if( $section_id !== 'link_type' ) {

			$control = $element->get_controls( $section_id );
			$control['conditions'] = [
				'terms' => [
					[
						'name' 		=> 'render_type',
						'operator'=> 'in',
						'value' 	=> ['custom_button'],
					],
				]
			];

			$element->update_control( $section_id, $control );
		}

		if( $section_id !== 'section_button' ) return;

		// We hide the link control, we will set it according to the lesson 
		utils\hide_control( $element, 'link' );
    utils\hide_control( $element, 'text' );

    // Mark complete ID -> needed for the video progression compatibility
 		utils\hide_control( $element, 'button_css_id', 'llms_mark_complete' );

		$element->set_render();
	}


	/**
	 * Sets the render.
	 *
	 * @param      obj  $widget  The widget
	 */
	public function set_render() {
		
		$widget = &$this;

		add_action('elementor/widget/render_content', function( $content, $element ) use( $widget ) {
		  	
	  	if( $element->get_name() === $widget->get_name() ) {
	  			
  			$settings = $element->get_settings_for_display();
        $lesson = lift\lesson_lifter();
  			$lifter_render = $this->lifter_render();

        if( empty($lifter_render) || !lift\is_lesson_drip_past( $lesson ) ) $content = '';

				/**
				 * If render_type is set on lifter_element 
				 * we bypass the render defiened in the parent class
				 */
				else if( $settings['render_type'] === 'lifter' ) {
					$content = $lifter_render;
				}
				/**
				 * Re-run render with the new URL
				 */
				else{

					if( !is_object($lesson) ) return $content;

          // False if none 
          $assignment_url = lift\lesson_assignment_url();

					if( \llms_show_take_quiz_button($lesson) || $assignment_url ) { 

            // In some lesson, it's possible to have both assignment and quizz button
            $quizz = '';
            $assignment = '';

            if( \llms_show_take_quiz_button($lesson) ) {
              $quizz = $content;
              $quizz = utils\change_button_text( $quizz, $settings['text_quiz'] );
						  $quizz = utils\modify_link( $quizz, lift\lesson_quiz_url() );
						  $quizz = "<div class='le-button-mark-quiz'>{$quizz}</div>"; 
            }

            if( $assignment_url ) {
              $assignment = $content;
						  $assignment = utils\change_button_text( $assignment, $settings['text_assignment'] );
						  $assignment = utils\modify_link( $assignment, $assignment_url );
						  $assignment = "<div class='le-button-mark-assignment'>{$assignment}</div>"; 
            }

            $content = $assignment . $quizz; 
					}
					else if( lift\can_retake_lesson() && lift\has_complete_lesson() ) {
						$content = utils\change_button_text( $content, $settings['text_incomplete'] );
						$content = $this->replace_link_by_button( $content, false );
						$content = "<div class='le-button-mark-incomplete'>{$content}</div>"; 
						$content = lift\mark_complete_form( $content ); 
					}
					else{
            $content = utils\change_button_text( $content, $settings[ 'text_complete' ] );
            $content = $this->replace_link_by_button( $content );
            $content = "<div class='le-button-mark-complete'>{$content}</div>";
            $content = lift\mark_complete_form( $content);
					}
				}
	  	}

	  	return $content;
		}, 999, 2 );
	}


	/**
	 * Front-end render
	 * 
	 * @return string
	 */
	protected function lifter_render() {

		ob_start();
		echo do_shortcode("[lifterlms_lesson_mark_complete]");
		return ob_get_clean();
	}


	/**
	 * Editor output for generating the live preview
	 */
	protected function content_template() {
		$this->set_render();
	}


	/**
	 * Remplace a <a> tag by a <button>
	 *
	 * @param      string  	$content  	The content (HTML)
	 * @param      boolean  $complete  	True for complete false for incomplete (HTML)
	 *
	 * @return     string
	 */
	protected function replace_link_by_button( $content, $complete = true ) {

		$doc = new \DOMDocument();
		@$doc->loadHTML( '<?xml encoding="utf-8" ?>' . $content );

		$links = $doc->getElementsByTagName('a');

		// Remplace link by button (keep the class)
		foreach( $links as $key => $link ) {

			// Create the button
			$button = $doc->createElement( "button" );
    	$button->setAttribute( 'class', $link->getAttribute( 'class' ) );
    	$button->setAttribute( 'id', $link->getAttribute( 'id' ) );
    	$button->setAttribute( 'type', 'submit' );

    	if( $complete ) {
    		$button->setAttribute( 'name', 'mark_complete' );
      }
      else{
        $button->setAttribute( 'name', 'mark_incomplete' );
      }
      
      // We move the content of the link to the button
      while( $link->hasChildNodes() ) {
        $button->appendChild( $link->firstChild );
      }

      // Remplace the link by the button
      $link->parentNode->replaceChild( $button, $link );
    }

    return "<div class='elementor-widget-button' >" . utils\get_html( $doc ) . "</div>";
  }
}


