<?php

namespace BetterLifter;

defined('ABSPATH') or die();

use BetterLifter as lift;


/**
 * Return the instance of the access plans of the 
 * current course 
 * 
 * @see plugins/lifterlms/includes/models/model.llms.access.plan.php
 * @see plugins/lifterlms/includes/models/model.llms.product.php
 * 
 * @param      integer  $course_id  The course identifier
 *
 * @return     array    
 */
function course_plans( $course_id = 0 ) {

  $course = lift\course_lifter( $course_id );
  
  if( empty($course) || !is_object($course) ) return array();
  
  $plans = lift\plans();
  $product = new \LLMS_Product( $course->get('id') );

  if( empty($product) || !is_object($product) ) return array();

  $response = array();

  foreach( $product->get_access_plans( false, false ) as $plan ) {
    array_push($response, $plan);
  }
    
  return $response;
}


/**
 * Return the IDs of the access plans of the 
 * current course 
 * 
 * @see plugins/lifterlms/includes/models/model.llms.access.plan.php
 * @see plugins/lifterlms/includes/models/model.llms.product.php
 * 
 * @param      integer  $course_id  The course identifier
 *
 * @return     array   	
 */
function course_plan_ids( $course_id = 0 ) {
  
  $course = lift\course_lifter( $course_id );

  if( empty($course) || !is_object($course) ) return array();
		
	$plans = lift\plans();
	$product = new \LLMS_Product( $course->get('id') );

	if( empty($product) || !is_object($product) ) return array();

	$response = array();

	foreach( $product->get_access_plans( false, false ) as $plan ) {
		array_push($response, $plan->get('id'));
	}
  	
  return $response;
}


/**
 * Return the IDs of the featured access plans of the 
 * current course 
 * 
 * @see plugins/lifterlms/includes/models/model.llms.access.plan.php
 * @see plugins/lifterlms/includes/models/model.llms.product.php
 * 
 * @param      integer  $course_id  The course identifier
 *
 * @return     array   	
 */
function course_featured_plan_ids( $course_id = 0 ) {
  
  $plans = lift\course_plan_ids( $course_id );

  if( empty($plans) ) return array();

	$response = array();

	foreach( $plans as $plan_id ) {
		$plan = new \LLMS_Access_Plan( $plan_id );
		if( $plan->is_featured() ) array_push($response, $plan->get('id'));
	}
  	
  return $response;
}


/**
 * Return the first access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     int
 */
function course_plan_id( $course_id = 0 ) {
  
  $plans = lift\course_plan_ids( $course_id );

  if( empty($plans) ) return '';
  	
  return $plans[0];
}


/**
 * Return the first featured access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     int
 */
function course_featured_plan_id( $course_id = 0 ) {
  
  $plans = lift\course_featured_plan_ids( $course_id );

  if( empty($plans) ) return '';
  	
  return $plans[0];
}


/**
 * Return the name of the first access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string
 */
function course_plan_name( $course_id = 0 ) {
  
  $plan_id = lift\course_plan_id( $course_id );

  if( empty($plan_id) ) return '';
  	
  return \get_the_title( $plan_id );
}


/**
 * Return the name of the first featured access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string
 */
function course_featured_plan_name( $course_id = 0 ) {
  
  	$plan_id = lift\course_featured_plan_id( $course_id );

  	if( empty($plan_id) ) return '';
  	
  	return \get_the_title( $plan_id );
}


/**
 * Return the url of the first featured access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string   url
 */
function course_featured_plan_url( $course_id = 0 ) {

	$plan_id = course_featured_plan_id( $course_id );
	
	if( empty($plan_id) ) return '';

	return lift\plan_url( $plan_id );
}


/**
 * Return the url of the first access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string   url
 */
function course_plan_url( $course_id = 0 ) {
	
	$plan_id = course_plan_id( $course_id );
	
	if( empty($plan_id) ) return '';

	return lift\plan_url( $plan_id );
}


/**
 * Return the price of the first access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string   url
 */
function course_plan_price( $course_id = 0 ) {
	
	$plan_id = lift\course_plan_id( $course_id );
	
	if( empty($plan_id) ) return '';

  return !lift\is_plan_on_sale( $plan_id ) ?
    lift\plan_price( $plan_id ) :
    lift\currency_symbol() . lift\plan_price_number_int( $plan_id );
}


/**
 * Return the price of the first featured access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string   url
 */
function course_featured_plan_price( $course_id = 0 ) {

	$plan_id = course_featured_plan_id( $course_id );
	
	if( empty($plan_id) ) return '';

  return !lift\is_plan_on_sale( $plan_id ) ?
    lift\plan_price( $plan_id ) :
    lift\currency_symbol() . lift\plan_price_number_int( $plan_id );
}


/**
 * Return the frequency of the first access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string   url
 */
function course_plan_frequency( $course_id = 0 ) {
	
	$plan_id = course_plan_id( $course_id );
	
	if( empty($plan_id) ) return '';

	return lift\plan_frequency( $plan_id );
}


/**
 * Return the frequency of the first featured access plan of the course (according 
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string   url
 */
function course_featured_plan_frequency( $course_id = 0 ) {

	$plan_id = course_featured_plan_id( $course_id );
	
	if( empty($plan_id) ) return '';

	return lift\plan_frequency( $plan_id );
}


/**
 * Return the enroll text of the first featured acces plan of the course (according
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string   ( description_of_the_return_value )
 */
function course_featured_plan_enroll_text( $course_id = 0 ) {
	
	$plan_id = course_featured_plan_id( $course_id );
	
	if( empty($plan_id) ) return '';

	return lift\plan_enroll_text( $plan_id );
}


/**
 * Return the enroll text of the first acces plan of the course (according
 * to the order in the drag and drop of the admin course page)
 *
 * @param      integer  $course_id  The course identifier
 *
 * @return     string   ( description_of_the_return_value )
 */
function course_plan_enroll_text( $course_id = 0 ) {
	
	$plan_id = course_plan_id( $course_id );
	
	if( empty($plan_id) ) return '';

	return lift\plan_enroll_text( $plan_id );
}
