<?php

namespace BetterLifter;

defined('ABSPATH') or die();

use BetterLifter as lift;


function course_post_type() {
  return 'course';
}


function course( $post_id = 0 ) {
  $course_id = lift\course_id($post_id);
  if( !empty($course_id) ) return get_post($course_id);
}


/**
 * Extract the text paragraph from the content
 *
 * @see https://github.com/WordPress/gutenberg/issues/6358 
 * 
 * @param      integer  $post_id  The post identifier
 *
 * @return     string
 */
function course_content( $post_id = 0 ) {
  
  $course_id = lift\course_id( $post_id ); 
  if( empty($course_id) ) return ''; 

  // Get the Lifter related Gutenberg block names
  $lifter_blocks = lift\lifter_block_names();
  $post = get_post( $course_id ); 

  if( has_blocks( $post->post_content ) ) {

    $blocks = parse_blocks( $post->post_content );
    $content = '';

    // Echo only blocks not related to LifterLMS
    foreach( $blocks as $block ) { 
      if( !in_array($block['blockName'], $lifter_blocks) ) {
        $content .= \render_block($block);
      }
    }

    /**
     * LifterLMS adds HTML we don't want in 'the_content' filter
     * 
     * @see lifterlms/includes/functions/llms-functions-content.php 
     */
    remove_filter( 'the_content', 'llms_get_post_content' );

    /**
     * We have to use 'the_content' filter for completly display embed-core blocks (youtube, vimeo ...)
     *
     * @see https://github.com/WordPress/gutenberg/issues/14080#issuecomment-567429414
     */
    $content = apply_filters( 'the_content', $content );
    
    add_filter( 'the_content', 'llms_get_post_content' );
  }
  // If there is not block, the content come from the classic editor and we can display as it is
  else{
    $content = $post->post_content; 
  }

  return do_shortcode($content);
}


/**
 * Return a lifterLMS Course instance
 *
 * @param      integer  $post_id  The post identifier
 *
 * @return     obj
 */
function course_lifter( $post_id = 0 ) {
  $course_id = lift\course_id( $post_id );
  if( !empty($course_id) ) return new \LLMS_Course( $course_id );
}

  
/**
 * Get the course excerpt
 *
 * @param      integer  $post_id  The post identifier
 * @param      integer  $character_number  The character number
 * 
 * @return     string
 */
function course_excerpt( $post_id = 0, $character_number = 55, $excerpt_more = '...' ) {

  $content = lift\course_content( $post_id );
  if( empty($content) ) return ''; 

  /**
   * @see wordpress function wp_trim_excerpt()
   */
  $content = strip_shortcodes( $content );
  $content = excerpt_remove_blocks( $content );
  $content = str_replace( ']]>', ']]&gt;', $content );
  
  return \wp_trim_words( $content, $character_number, $excerpt_more );
}


/**
 * Determines whether the specified post identifier is course.
 *
 * @param      integer  $post_id  The post identifier
 *
 * @return     boolean  True if the specified post identifier is course, False otherwise.
 */
function is_course( $post_id = 0 ) {
  if( empty($post_id) ) $post_id = get_the_id();
  $post = get_post($post_id);
  return !empty($post) && $post->post_type === lift\course_post_type();
}


/**
 * @see lifterlms/includes/models/model.llms.lesson.php
 */
function course_id( $post_id = 0 ) {

  if( $post_id instanceof \WP_Post ) $post_id = $post_id->ID;
  if( empty($post_id) ) $post_id = get_the_id();
  if( lift\is_course($post_id) ) return $post_id;
  if( lift\is_lesson($post_id) ) {
  	$lesson = new \LLMS_Lesson( $post_id );
  	return $lesson->get('parent_course'); 
  }
  
  return 0;
}


/**
 * Return the url of the course
 *
 * @param      integer  $post_id  The post identifier
 *
 * @return     string   url
 */
function course_url( $post_id = 0 ) {

  $course_id = lift\course_id( $post_id );
  return get_the_permalink( $course_id );
}


/**
 * Return the title of the course
 *
 * @param      integer  $post_id  The post identifier
 *
 * @return     string   
 */
function course_title( $post_id = 0 ) {

  $course_id = lift\course_id( $post_id );
  return get_the_title( $course_id );
}
