<?php

namespace BetterLifter;

defined('ABSPATH') or die();

use BetterLifter as lift;

/**
 * Drip time expired or not
 * @return  boolean
 */
function is_lesson_drip_past( $lesson ) {

  $expired = true;
  $drip_time = lift\lesson_drip_time( $lesson );

  // It is also 'expired' for Users who have 'Unrestricted Preview Access' set in LifterLMS 'General Settings'
	if ( $drip_time == 0 || llms_can_user_bypass_restrictions() ) return $expired;

  $diff = $drip_time - time();
  $expired = $diff <= 0;

  return $expired;
}

/**
 * Drip timestamp if exists, otherwise 0
 * @return mixed
 */
function lesson_drip_time( $lesson ) {


  $drip_time = 0; // Force past date
  
  //Pass $post_id, otherwise won't work properly if called from lessons archive or if called from a page of another post type - 'is_lesson' always returns false in those cases
  $post_id = ($lesson instanceof \LLMS_Lesson) ? $lesson -> id : 0;
  if (!lift\is_lesson($post_id)) return $drip_time;
  
  $drip_method = $lesson->drip_method;
  if (!$lesson->drip_method) return $drip_time; //  '', date, enrollment, prerequisite

  $seconds_per_day = 86400;

  if ( $drip_method == 'date' ) {

    $date_available = $lesson->date_available;
    $time_available = $lesson->time_available;
    $available_date = new \DateTime( $date_available.' '.$time_available );

    $drip_time = $available_date->getTimestamp();

  } elseif ( $drip_method == 'start' ) {

    $course = lift\course_lifter( $lesson->parent_course );
    $time_period_enabled =  ( 'yes' === $course -> get( 'time_period' ) ) ? true : false ;

    if ( $time_period_enabled ) {

      $course_start_date = $course -> get_date( 'start_date', 'Y-m-d\TH:i:s\Z' );
      if( !$course_start_date ) return $drip_time;

      $start_date = \DateTime::createFromFormat( 
        'Y-m-d\TH:i:s\Z', 
        $course_start_date 
      );
      $start_time = $start_date -> getTimestamp();

      $before_available_time = $lesson->days_before_available * $seconds_per_day;
      $drip_time = $start_time + $before_available_time;
    }

  } elseif ( $drip_method == 'enrollment' ) {
    
    $course = lift\course_lifter( $lesson->parent_course );
    $user = new \LLMS_Student();

    $is_enrolled = lift\user_course_is_enrolled( $lesson->parent_course );

    if ( $is_enrolled ) {
      
      $enrollment_date = \DateTime::createFromFormat( 
        'Y-m-d\TH:i:s\Z',
        $user->get_enrollment_date( $lesson->parent_course, 'updated', 'Y-m-d\TH:i:s\Z' )
      );
      $enrollment_time = $enrollment_date->getTimestamp();

      $before_available_time = $lesson->days_before_available * $seconds_per_day;
      $drip_time = $enrollment_time + $before_available_time;
    }

  } elseif ( $drip_method == 'prerequisite' && $lesson->has_prerequisite == 'yes' ) {

    $user = new \LLMS_Student();
    $prerequisite = lift\lesson_lifter( $lesson->prerequisite );
    $prerequisite_completed = lift\has_complete_lesson( $lesson->prerequisite );

    if ( $prerequisite_completed ) {

      $prerequisite_completed_date = \DateTime::createFromFormat(
        'Y-m-d\TH:i:s\Z',
        $user->get_completion_date( $lesson->prerequisite, 'Y-m-d\TH:i:s\Z' )
      );
      $prerequisite_completed_time =  $prerequisite_completed_date -> getTimestamp();

      $before_available_time = $lesson->days_before_available * $seconds_per_day;
      $drip_time = $prerequisite_completed_time + $before_available_time;
    }
  }

  return $drip_time;
}

/**
 * Return the html of the lesson timer
 *
 * @param      array   $atts   The atts
 *
 * @return    string|void
 */
function lesson_timer_html( $atts = array() ) {

    $lesson = lift\lesson_lifter();
    if ( !$lesson ) return;

    $a = shortcode_atts( array(
      'show' => 'all',
      'round' => 'no',
      'which_lesson' => 'current',
      'id' => 0,
    ), $atts );

    $which_lesson = $a['which_lesson'];
    $lesson = ( $which_lesson === 'current' ) ? $lesson : lift\lesson_lifter( $lesson -> get_next_lesson() );
    $expired = lift\is_lesson_drip_past( $lesson );
    if ( $expired ) return;

    $show = $a['show'];
    $round = $a['round'];
    $drip_time = lift\lesson_drip_time( $lesson );
    $id = $a['id'];

    $html = '';
    $html .= '<div 
        id="le-lesson-drip-timer-' . $id . '" 
        class="le-lesson-drip-timer-wrapper"  
        data-timer=\'{"show":"' . $show . '", "driptime":"' . $drip_time . '", "round":"' . $round . '", "id":"' . $id . '" }\' >';
    $html .= '<div class="days"></div>';
    $html .= '<div class="hours"></div>';
    $html .= '<div class="minutes"></div>';
    $html .= '<div class="seconds"></div>';
    $html .= '</div>';

    return $html;
}

/**
 * If Dripped Lesson view allowed true, else false
 * @return boolean
 */
function is_dripped_lesson_view_allowed( $lesson ) {

  if( !$lesson || !($lesson instanceof \LLMS_Lesson) ) return false;

  // plugin settings - drip view allowed for all lessons
  $settings_all = lift\get_setting( 'lessons', 'drip_view_all_lessons' );

  // if dripped lesson view is allowed for all lessons we finish here
  if ( $settings_all === 'true' )  return true;

  // plugin settings - dripped lesson view allowed for specific course categories
  $settings_categories = lift\get_setting( 'lessons', 'drip_view_course_categories' );

  // if there is no course categories set we finish here
  if (empty($settings_categories))  return false;

  // Get Lessons' parent Course
  $parent_course_id = $lesson -> get('parent_course');
  $course = lift\course_lifter( $parent_course_id );

  if( empty($course) ) return false;
  $course_categories = $course -> get_categories( ['fields' => 'ids' ] );

  // find if any of Course categories belongs to those which allow redirection - if not return false
  $matches = array_intersect( $settings_categories, $course_categories );
  if( !$matches ) return false;

  return true;
}
