<?php

namespace Tangible\ElementLogic;

defined( 'ABSPATH' ) or die();

use Tangible\ElementLogic\Control as Control;
use Tangible\Logic as logic;

/**
 * @see https://docs.tangible.one/modules/logic
 * @see https://docs.tangible.one/modules/logic/elementor
 * @see https://docs.tangible.one/modules/plugin-framework/plugin-groups
 */
if( ! class_exists( '\Tangible\ElementLogic\ConditionalLogic' ) ) {
  class ConditionalLogic {

    private $fields;
    private $callbacks;
    
    /**
     * Constructs a new instance.
     */
    public function __construct() {
      $this->register_group();
    }


    /**
     * Register group config and actions
     *
     * @see https://docs.tangible.one/modules/plugin-framework/plugin-groups
     */
    private function register_group() {
      // Every plugin in the group registers its rules/fields, and an "evaluate rules" callback
      Controller::$instance->state['group']->unique_action( 'register_logic_ui', [$this, 'register_logic_ui'] );
    }


    /**
     * Register shared logic UI and visibility actions for the whole group
     */
    public function register_logic_ui() {

      if( Controller::$instance->state['initialized'] ) return;

      // Avoid to call this function multiple time 
      Controller::$instance->state['initialized'] = true;

      $config = Controller::$instance->state['group']->config;

      // Pass combined group fields to Control
      $this->fields = $config['fields'];

      // Pass evaluator callbacks to is_element_visible
      $this->callbacks = $config['callbacks'];

      // Widget content visibility
      add_action('elementor/widget/render_content', function($content, $element) {
        if( $this->is_element_visible($element) ) return $content;
        return '';
      }, 10, 2 );

      $this->register_control();

      // Register Logic control 
      add_action('elementor/init', function() {

        // Add to "Advanced" tab
        add_action('elementor/element/before_section_start', function($element, $section_id, $args) {

          if ('_section_responsive' !== $section_id) return;
          
          $element->start_controls_section(
            'logic_controls_section',
            [
              'tab'   => \Elementor\Controls_Manager::TAB_ADVANCED,
              'label' => 'Conditional Logic',
            ]
          );

          $element->add_control(
            'logic_control',
            [
              'label'       => 'Conditional Logic Data',
              'type'        => Control::FIELD_NAME,
              'render_type' => 'none',
            ]
          );

          $element->end_controls_section();

          // Add a message on the content table for noticing the users about the conditional logic (only for widgets)
          $tab_type = ! in_array($element->get_name(), ['section', 'column', 'container']) 
            ? \Elementor\Controls_Manager::TAB_CONTENT 
            : \Elementor\Controls_Manager::TAB_LAYOUT;  
          
          $element->start_controls_section(
            'logic_controls_note',
            [
              'tab'   => $tab_type,
              'label' => 'New!',
            ]
          );

          $element->add_control(
            'logic_controls_note_html',
            [
              'label' => '',
              'type'  => \Elementor\Controls_Manager::RAW_HTML,
              'raw'   => 'Check out the conditional logic panel in the advanced section!',
            ]
          );

          $element->end_controls_section();
          
        }, 10, 3);  
      });

      /**
       * We could probably symplify and only use the elementor/frontend/after_render action now that it exists
       * @see print_element() in elementor/includes/base/element-base.php
       */
      add_action( 'elementor/frontend/widget/before_render', [ $this, 'wrapper_visibility' ] );
      add_action( 'elementor/frontend/section/before_render', [ $this, 'wrapper_visibility' ] );
      add_action( 'elementor/frontend/column/before_render', [ $this, 'wrapper_visibility' ] );
      add_action( 'elementor/frontend/container/before_render', [ $this, 'wrapper_visibility'] );
    }

    /**
     * Register logic panel as a control in Elementor
     */
    public function register_control() {
      
      /**
       * @see https://developers.elementor.com/docs/managers/registering-controls/#registering-new-controls
       */
      add_action('elementor/controls/register', function($controls_manager) {
        
        $logic_panel = new Control($this->fields);

        $controls_manager->register( $logic_panel );

      });
      
    }

    /**
     * Evaluate elemeent's visibility
     *
     * @param      object  $element  The element
     */
    public function wrapper_visibility( $element ) {

      if( $this->is_element_visible($element) ) return;

      $element->add_render_attribute( '_wrapper', [
        'style' => 'display:none' ,
      ]);
    }


    /**
     * Evaluate the visibility of an elementor element
     * 
     * @param  obj  $element
     * 
     * @return boolean
     */
    public function is_element_visible( $element ) {

      $settings = $element->get_active_settings();
      if (empty($settings['logic_control'])) return true;

      $cache_key = '_logic_control_cached_result';
      if (isset($settings[$cache_key])) return $settings[$cache_key];

      // Rule groups saved from logic UI
      $rule_groups = json_decode($settings['logic_control']);

      // Callbacks to evaluate rules for the group
      $callbacks = $this->callbacks;

      // Context - it can be anything needed for the evaluate rule function
      $context = get_post();

      $visibility = true;

      foreach ($callbacks as $callback) {
        $visibility = logic\evaluate_rule_groups($rule_groups, $callback, $context);
        // All rules must be satisfied for final visibility result
        if (!$visibility) break;
      }

      $element->set_settings($cache_key, $visibility);

      return $visibility;
    }
    
  }
}

